/*
 * Copyright © 2021-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel24-tictactoe-core.
 *
 * ipspiel24-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel24-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel24-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel24.tictactoe.core.domain.impl;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.PlayerState;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeBoard;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeField;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeFieldState;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToePlayer;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToePosition;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeRow;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeState;

/**
 * Implements {@link TicTacToeState}.
 */
@SuppressWarnings("PMD.GodClass")
final class TicTacToeStateImpl implements TicTacToeState {

    /**
     * The board.
     */
    private final TicTacToeBoard board;

    /**
     * The player using crosses.
     */
    private final TicTacToePlayer crossesPlayer;

    /**
     * The player using noughts.
     */
    private final TicTacToePlayer noughtsPlayer;
    
    /**
     * The current player.
     */
    private TicTacToePlayer currentPlayer;

    /**
     * Creates an Tic Tac Toe state.
     *
     * @param board         The board.
     * @param crossesPlayer The player using crosses.
     * @param noughtsPlayer The player using noughts.
     * @param crossesIsNext {@code true} if the player using crosses is next, else the player using noughts is next.
     * @throws GameException if the state cannot be created according to the rules of the game.
     */
    TicTacToeStateImpl(final TicTacToeBoard board, final TicTacToePlayer crossesPlayer,
            final TicTacToePlayer noughtsPlayer, final boolean crossesIsNext) throws GameException {

        this.board = Objects.requireNonNull(board, "board");
        this.crossesPlayer = Objects.requireNonNull(crossesPlayer, "crossesPlayer");
        this.noughtsPlayer = Objects.requireNonNull(noughtsPlayer, "noughtsPlayer");
        this.currentPlayer = crossesIsNext ? this.crossesPlayer : this.noughtsPlayer;

        if (!this.crossesPlayer.isUsingCrosses()) {
            throw new IllegalArgumentException(String.format("Player %s does not use crosses.", this.crossesPlayer));
        }
        if (this.noughtsPlayer.isUsingCrosses()) {
            throw new IllegalArgumentException(String.format("Player %s does not use noughts.", this.noughtsPlayer));
        }
        if (this.crossesPlayer.getName().equals(this.noughtsPlayer.getName())) {
            throw new IllegalArgumentException(
                    String.format("Both players have the same name '%s'.", this.crossesPlayer.getName()));
        }
    }

    /**
     * Creates an TicTacToe state by copying an existing one.
     *
     * @param source The state to copy.
     */
    TicTacToeStateImpl(final TicTacToeStateImpl source) {
        this.board = source.board.deepCopy();
        this.crossesPlayer = source.crossesPlayer.deepCopy();
        this.noughtsPlayer = source.noughtsPlayer.deepCopy();
        this.currentPlayer = source.isCrossesPlayerCurrent() ? this.crossesPlayer : this.noughtsPlayer;
    }

    @Override
    public String toString() {
        return String.format(
                "TicTacToeState[board=\n%s, crossesPlayer=%s, noughtsPlayer=%s, currentPlayer=%s]",
                this.board,
                this.crossesPlayer,
                this.noughtsPlayer,
                this.currentPlayer.isUsingCrosses() ? TicTacToeFieldState.CROSS : TicTacToeFieldState.NOUGHT);
    }

    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof TicTacToeStateImpl) {
            final TicTacToeStateImpl other = (TicTacToeStateImpl) obj;
            return this.board.equals(other.board) && this.crossesPlayer.equals(other.crossesPlayer)
                    && this.noughtsPlayer.equals(other.noughtsPlayer)
                    && this.isCrossesPlayerCurrent() == other.isCrossesPlayerCurrent();
        }
        return false;
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.board, this.crossesPlayer, this.noughtsPlayer, this.isCrossesPlayerCurrent());
    }

    @Override
    public TicTacToeStateImpl deepCopy() {
        return new TicTacToeStateImpl(this);
    }

    @Override
    public Map<String, TicTacToePlayer> getPlayers() {
        final Map<String, TicTacToePlayer> result = new LinkedHashMap<>();
        result.put(this.crossesPlayer.getName(), this.crossesPlayer);
        result.put(this.noughtsPlayer.getName(), this.noughtsPlayer);
        return result;
    }

    @Override
    public TicTacToeBoard getBoard() {
        return this.board;
    }

    @Override
    public TicTacToePlayer getCrossesPlayer() {
        return this.crossesPlayer;
    }

    @Override
    public TicTacToePlayer getNoughtsPlayer() {
        return this.noughtsPlayer;
    }

    @Override
    public TicTacToePlayer getCurrentPlayer() {
        return this.currentPlayer;
    }

    @Override
    public void moveCompleted() {
        final Set<TicTacToeRow> uniformRows = this.board.getRowsUniformlyMarked();
        if (!uniformRows.isEmpty()) {
            final TicTacToeFieldState rowState = uniformRows.iterator().next().getState().orElseThrow();
            this.gameOver(rowState);
        } else {
            final Map<TicTacToePosition,
                    ? extends TicTacToeField> emptyFields = this.board.getFieldsBeing(TicTacToeFieldState.EMPTY);
            if (emptyFields.isEmpty()) {
                this.gameOver(TicTacToeFieldState.EMPTY);
            }
        }
    }

    @Override
    public Set<TicTacToePlayer> computeNextPlayers() {
        if (this.currentPlayer.getState().equals(PlayerState.PLAYING)) {
            return Collections.singleton(this.currentPlayer);
        }
        return Collections.emptySet();
    }

    @Override
    public void nextTurn() {
        this.currentPlayer = this.getOtherPlayer();
    }

    /**
     * Returns the currently inactive player.
     */
    private TicTacToePlayer getOtherPlayer() {
        if (this.isCrossesPlayerCurrent()) {
            return this.noughtsPlayer;
        }
        return this.crossesPlayer;
    }

    /**
     * Returns {@code true} iff the current player uses crosses.
     */
    @SuppressWarnings("PMD.CompareObjectsWithEquals")
    private boolean isCrossesPlayerCurrent() {
        return this.currentPlayer == this.crossesPlayer;
    }

    /**
     * Updates the player states after the game has finished.
     *
     * @param rowState The state of some uniformly marked row, or {@link TicTacToeFieldState#EMPTY} if no player has won
     *                 the game yet and no empty fields have remained.
     */
    private void gameOver(final TicTacToeFieldState rowState) {
        switch (rowState) {
        case CROSS:
            this.crossesPlayer.setState(PlayerState.WON);
            this.noughtsPlayer.setState(PlayerState.LOST);
            break;
        case NOUGHT:
            this.crossesPlayer.setState(PlayerState.LOST);
            this.noughtsPlayer.setState(PlayerState.WON);
            break;
        case EMPTY:
            this.crossesPlayer.setState(PlayerState.DRAW);
            this.noughtsPlayer.setState(PlayerState.DRAW);
            break;
        default:
            throw new IllegalArgumentException(
                    String.format("TicTacToeState.gameOver() called with state %s.", rowState));
        }
    }
}
